@@ -85,13 +85,6 @@ NMOpenvpnPlugin *nm_openvpn_plugin_new (const char *bus_name);
 
 /*****************************************************************************/
 
-typedef enum {
-	OPENVPN_BINARY_VERSION_INVALID,
-	OPENVPN_BINARY_VERSION_UNKNOWN,
-	OPENVPN_BINARY_VERSION_2_3_OR_OLDER,
-	OPENVPN_BINARY_VERSION_2_4_OR_NEWER,
-} OpenvpnBinaryVersion;
-
 typedef struct {
 	GPid pid;
 	guint watch_id;
@@ -110,6 +103,8 @@ typedef struct {
 	char *pending_auth;
 	char *challenge_state_id;
 	char *challenge_text;
+	char *challenge_response;
+	char *challenge_flags;
 	GIOChannel *socket_channel;
 	guint socket_channel_eventid;
 } NMOpenvpnPluginIOData;
@@ -138,6 +133,7 @@ typedef struct {
 } ValidProperty;
 
 static const ValidProperty valid_properties[] = {
+	{ NM_OPENVPN_KEY_ALLOW_COMPRESSION,         G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_ALLOW_PULL_FQDN,           G_TYPE_BOOLEAN, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_AUTH,                      G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_CA,                        G_TYPE_STRING, 0, 0, FALSE },
@@ -150,6 +146,7 @@ static const ValidProperty valid_properties[] = {
 	{ NM_OPENVPN_KEY_CONNECTION_TYPE,           G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_CRL_VERIFY_FILE,           G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_CRL_VERIFY_DIR,            G_TYPE_STRING, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_DATA_CIPHERS,              G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_EXTRA_CERTS,               G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_FLOAT,                     G_TYPE_BOOLEAN, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_NCP_DISABLE,               G_TYPE_BOOLEAN, 0, 0, FALSE },
@@ -197,7 +194,9 @@ static const ValidProperty valid_properties[] = {
 	{ NM_OPENVPN_KEY_NOSECRET,                  G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_HTTP_PROXY_PASSWORD_FLAGS, G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_TLS_VERSION_MIN,           G_TYPE_STRING, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_TLS_VERSION_MIN_OR_HIGHEST,G_TYPE_BOOLEAN, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_TLS_VERSION_MAX,           G_TYPE_STRING, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_CHALLENGE_RESPONSE_FLAGS,  G_TYPE_STRING, 0, 0, FALSE },
 	{ NULL,                                     G_TYPE_NONE, FALSE }
 };
 
@@ -206,6 +205,7 @@ static const ValidProperty valid_secrets[] = {
 	{ NM_OPENVPN_KEY_CERTPASS,             G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_NOSECRET,             G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_HTTP_PROXY_PASSWORD,  G_TYPE_STRING, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_CHALLENGE_RESPONSE,   G_TYPE_STRING, 0, 0, FALSE },
 	{ NULL,                                G_TYPE_NONE, FALSE }
 };
 
@@ -364,8 +364,10 @@ args_add_vpn_certs (GPtrArray *args, NMSettingVpn *s_vpn)
 	if (nmovpn_arg_is_set (cert) && is_pkcs12 (cert))
 		args_add_strv (args, "--pkcs12", cert);
 	else {
-		args_add_strv (args, "--cert", cert);
-		args_add_strv (args, "--key", key);
+		if (nmovpn_arg_is_set (cert))
+			args_add_strv (args, "--cert", cert);
+		if (nmovpn_arg_is_set (key))
+			args_add_strv (args, "--key", key);
 	}
 }
 
@@ -525,15 +527,13 @@ openvpn_binary_find_exepath (void)
 	return NULL;
 }
 
-static OpenvpnBinaryVersion
+static guint
 openvpn_binary_detect_version (const char *exepath)
 {
 	gs_free char *s_stdout = NULL;
-	const char *s;
 	int exit_code;
-	int n;
 
-	g_return_val_if_fail (exepath && exepath[0] == '/', OPENVPN_BINARY_VERSION_UNKNOWN);
+	g_return_val_if_fail (exepath && exepath[0] == '/', NMOVPN_VERSION_UNKNOWN);
 
 	if (!g_spawn_sync (NULL,
 	                   (char *[]) { (char *) exepath, "--version", NULL },
@@ -545,58 +545,40 @@ openvpn_binary_detect_version (const char *exepath)
 	                   NULL,
 	                   &exit_code,
 	                   NULL))
-		return OPENVPN_BINARY_VERSION_UNKNOWN;
+		return NMOVPN_VERSION_UNKNOWN;
 
 	if (   !WIFEXITED (exit_code)
 	    || !NM_IN_SET(WEXITSTATUS (exit_code), 0, 1)) {
 		/* expect return code 1 (OPENVPN_EXIT_STATUS_USAGE).
 		 * Since 2.5.0, it returns 0. */
-		return OPENVPN_BINARY_VERSION_UNKNOWN;
+		return NMOVPN_VERSION_UNKNOWN;
 	}
 
-	/* the output for --version starts with title_string, which starts with PACKAGE_STRING,
-	 * which looks like "OpenVPN 2.#...". Do a strict parsing here... */
-	if (   !s_stdout
-	    || !g_str_has_prefix (s_stdout, "OpenVPN 2."))
-		return OPENVPN_BINARY_VERSION_UNKNOWN;
-	s = &s_stdout[NM_STRLEN ("OpenVPN 2.")];
-
-	if (!g_ascii_isdigit (s[0]))
-		return OPENVPN_BINARY_VERSION_UNKNOWN;
-
-	n = 0;
-	do {
-		if (n > G_MAXINT / 100)
-			return OPENVPN_BINARY_VERSION_UNKNOWN;
-		n = (n * 10) + (s[0] - '0');
-	} while (g_ascii_isdigit ((++s)[0]));
-
-	if (n <= 3)
-		return OPENVPN_BINARY_VERSION_2_3_OR_OLDER;
-	return OPENVPN_BINARY_VERSION_2_4_OR_NEWER;
+	return nmovpn_version_parse (s_stdout);
 }
 
-static OpenvpnBinaryVersion
-openvpn_binary_detect_version_cached (const char *exepath, OpenvpnBinaryVersion *cached)
+static guint
+openvpn_binary_detect_version_cached (const char *exepath, guint *cached)
 {
-	if (G_UNLIKELY (*cached == OPENVPN_BINARY_VERSION_INVALID)) {
-		const char *str;
+	guint v;
 
-		*cached = openvpn_binary_detect_version (exepath);
-		switch (*cached) {
-		case OPENVPN_BINARY_VERSION_2_3_OR_OLDER:
-			str = "2.3 or older";
-			break;
-		case OPENVPN_BINARY_VERSION_2_4_OR_NEWER:
-			str = "2.4 or newer";
-			break;
-		default:
-			str = "unknown";
-			break;
+	v = *cached;
+	if (G_UNLIKELY (v == NMOVPN_VERSION_INVALID)) {
+		v = openvpn_binary_detect_version (exepath);
+		if (v >= NMOVPN_VERSION_UNKNOWN) {
+			v = NMOVPN_VERSION_UNKNOWN;
+			_LOGI ("detected openvpn version UNKNOWN, assume max");
+		} else {
+			guint v_x;
+			guint v_y;
+			guint v_z;
+
+			nmovpn_version_decode (v, &v_x, &v_y, &v_z);
+			_LOGI ("detected openvpn version %u.%u.%u", v_x, v_y, v_z);
 		}
-		_LOGI ("detected openvpn version %s", str);
+		*cached = v;
 	}
-	return *cached;
+	return v;
 }
 
 /*****************************************************************************/
@@ -837,7 +819,7 @@ get_detail (const char *input, const char *prefix)
  * CRV1:flags:state_id:username:text
  */
 static gboolean
-parse_challenge (const char *failure_reason, char **challenge_state_id, char **challenge_text)
+parse_challenge (const char *failure_reason, char **challenge_state_id, char **challenge_text, char **challenge_flags)
 {
 	const char *colon[4];
 
@@ -861,6 +843,7 @@ parse_challenge (const char *failure_reason, char **challenge_state_id, char **c
 	if (!colon[3])
 		return FALSE;
 
+	*challenge_flags = g_strndup (colon[0] + 1, colon[1] - colon[0] - 1);
 	*challenge_state_id = g_strndup (colon[1] + 1, colon[2] - colon[1] - 1);
 	*challenge_text = g_strdup (colon[3] + 1);
 	return TRUE;
@@ -914,12 +897,12 @@ handle_auth (NMOpenvpnPluginIOData *io_data,
 		if (!username)
 			username = io_data->default_username;
 
-		if (username != NULL && io_data->password != NULL && io_data->challenge_state_id) {
+		if (username != NULL && io_data->challenge_response) {
 			gs_free char *response = NULL;
 
 			response = g_strdup_printf ("CRV1::%s::%s",
 			                            io_data->challenge_state_id,
-			                            io_data->password);
+			                            io_data->challenge_response);
 			write_user_pass (io_data->socket_channel,
 			                 requested_auth,
 			                 username,
@@ -937,14 +920,21 @@ handle_auth (NMOpenvpnPluginIOData *io_data,
 				hints[i++] = NM_OPENVPN_KEY_USERNAME;
 				*out_message = _("A username is required.");
 			}
-			if (!io_data->password) {
-				hints[i++] = NM_OPENVPN_KEY_PASSWORD;
-				*out_message = _("A password is required.");
-			}
-			if (!username && !io_data->password)
-				*out_message = _("A username and password are required.");
-			if (io_data->challenge_text)
+
+			if (io_data->challenge_state_id) {
+				/* If we have a challenge we must have already authenticated with a password */
+				if (strstr (io_data->challenge_flags, "E"))
+					hints[i++] = NM_OPENVPN_HINT_CHALLENGE_RESPONSE_ECHO;
+				else
+					hints[i++] = NM_OPENVPN_HINT_CHALLENGE_RESPONSE_NOECHO;
 				*out_message = io_data->challenge_text;
+			} else if (!io_data->password) {
+				hints[i++] = NM_OPENVPN_KEY_PASSWORD;
+				if (username)
+					*out_message = _ ("A password is required.");
+				else
+					*out_message = _ ("A username and password are required.");
+			}
 		}
 		handled = TRUE;
 	} else if (nm_streq (requested_auth, "Private Key")) {
@@ -1068,10 +1058,12 @@ handle_management_socket (NMOpenvpnPlugin *plugin,
 			gs_free char *failure_reason = NULL;
 
 			failure_reason = get_detail (str, ">PASSWORD:Verification Failed: 'Auth' ['");
-			if (parse_challenge (failure_reason, &priv->io_data->challenge_state_id, &priv->io_data->challenge_text)) {
-				_LOGD ("Received challenge '%s' for state '%s'",
+			if (parse_challenge (failure_reason, &priv->io_data->challenge_state_id,
+			                     &priv->io_data->challenge_text, &priv->io_data->challenge_flags)) {
+				_LOGD ("Received challenge '%s' for state '%s' with flags '%s'",
 				       priv->io_data->challenge_state_id,
-				       priv->io_data->challenge_text);
+				       priv->io_data->challenge_text,
+				       priv->io_data->challenge_flags);
 			} else
 				_LOGW ("Password verification failed");
 
@@ -1231,6 +1223,11 @@ update_io_data_from_vpn_setting (NMOpenvpnPluginIOData *io_data,
 		g_free (io_data->password);
 	}
 	io_data->password = g_strdup (nm_setting_vpn_get_secret (s_vpn, NM_OPENVPN_KEY_PASSWORD));
+	if (io_data->challenge_response) {
+		memset (io_data->challenge_response, 0, strlen (io_data->challenge_response));
+		g_free (io_data->challenge_response);
+	}
+	io_data->challenge_response = g_strdup (nm_setting_vpn_get_secret (s_vpn, NM_OPENVPN_KEY_CHALLENGE_RESPONSE));
 
 	if (io_data->priv_key_pass) {
 		memset (io_data->priv_key_pass, 0, strlen (io_data->priv_key_pass));
@@ -1336,6 +1333,7 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 	GPid pid;
 	gboolean dev_type_is_tap;
 	const char *defport, *proto_tcp;
+	const char *allow_compression = NULL;
 	const char *compress;
 	const char *tls_remote = NULL;
 	const char *nm_openvpn_user, *nm_openvpn_group, *nm_openvpn_chroot;
@@ -1343,10 +1341,11 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 	NMSettingVpn *s_vpn;
 	const char *connection_type;
 	gint64 v_int64;
-	OpenvpnBinaryVersion openvpn_binary_version = OPENVPN_BINARY_VERSION_INVALID;
+	guint openvpn_binary_version = NMOVPN_VERSION_INVALID;
 	guint num_remotes = 0;
 	gs_free char *cmd_log = NULL;
 	NMOvpnComp comp;
+	NMOvpnAllowCompression allow_comp;
 
 	s_vpn = nm_connection_get_setting_vpn (connection);
 	if (!s_vpn) {
@@ -1521,6 +1520,9 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 	 * See bgo#769177
 	 */
 
+	/* New (2.5+) allow-compression option ("yes", "no", "asym") */
+	allow_compression = nm_setting_vpn_get_data_item (s_vpn,
+	                                                  NM_OPENVPN_KEY_ALLOW_COMPRESSION);
 	/* New (2.4+) compress option ("lz4", "lzo", ...) */
 	compress = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_COMPRESS);
 	/* Legacy option ("yes", "adaptive", "no", ...) */
@@ -1529,43 +1531,53 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 	if (compress && tmp)
 		_LOGW ("'compress' option overrides 'comp-lzo'");
 
+	allow_comp = nmovpn_allow_compression_from_options (allow_compression);
 	comp = nmovpn_compression_from_options (tmp, compress);
 	openvpn_binary_detect_version_cached (openvpn_binary, &openvpn_binary_version);
 
-	switch (comp) {
-	case NMOVPN_COMP_DISABLED:
-		break;
-	case NMOVPN_COMP_LZO:
-		if (openvpn_binary_version == OPENVPN_BINARY_VERSION_2_3_OR_OLDER)
-			args_add_strv (args, "--comp-lzo", "yes");
-		else
-			args_add_strv (args, "--compress", "lzo");
-		break;
-	case NMOVPN_COMP_LZ4:
-	case NMOVPN_COMP_LZ4_V2:
-	case NMOVPN_COMP_AUTO:
-		if (openvpn_binary_version == OPENVPN_BINARY_VERSION_2_3_OR_OLDER)
-			_LOGW ("\"compress\" option supported only by OpenVPN >= 2.4");
-
-		if (comp == NMOVPN_COMP_LZ4)
-			args_add_strv (args, "--compress", "lz4");
-		else if (comp == NMOVPN_COMP_LZ4_V2)
-			args_add_strv (args, "--compress", "lz4-v2");
-		else
-			args_add_strv (args, "--compress");
-		break;
-	case NMOVPN_COMP_LEGACY_LZO_DISABLED:
-	case NMOVPN_COMP_LEGACY_LZO_ADAPTIVE:
-		if (openvpn_binary_version != OPENVPN_BINARY_VERSION_2_3_OR_OLDER)
-			_LOGW ("\"comp-lzo\" is deprecated and will be removed in future OpenVPN releases");
-
-		args_add_strv (args, "--comp-lzo",
-		                  comp == NMOVPN_COMP_LEGACY_LZO_DISABLED
-		               ? "no"
-		               : "adaptive");
-		break;
+	if (nmovpn_arg_is_set (allow_compression)) {
+		if (openvpn_binary_version < nmovpn_version_encode (2, 5, 0)) {
+			_LOGW ("\"allow-compression\" is only supported in OpenVPN 2.5 and later versions");
+		} else {
+			args_add_strv (args, "--allow-compression", allow_compression);
+		}
 	}
 
+	if (allow_comp != NMOVPN_ALLOW_COMPRESSION_NO)
+		switch (comp) {
+		case NMOVPN_COMP_DISABLED:
+			break;
+		case NMOVPN_COMP_LZO:
+			if (openvpn_binary_version < nmovpn_version_encode (2, 4, 0))
+				args_add_strv (args, "--comp-lzo", "yes");
+			else
+				args_add_strv (args, "--compress", "lzo");
+			break;
+		case NMOVPN_COMP_LZ4:
+		case NMOVPN_COMP_LZ4_V2:
+		case NMOVPN_COMP_AUTO:
+			if (openvpn_binary_version < nmovpn_version_encode (2, 4, 0))
+				_LOGW ("\"compress\" option supported only by OpenVPN >= 2.4");
+
+			if (comp == NMOVPN_COMP_LZ4)
+				args_add_strv (args, "--compress", "lz4");
+			else if (comp == NMOVPN_COMP_LZ4_V2)
+				args_add_strv (args, "--compress", "lz4-v2");
+			else
+				args_add_strv (args, "--compress");
+			break;
+		case NMOVPN_COMP_LEGACY_LZO_DISABLED:
+		case NMOVPN_COMP_LEGACY_LZO_ADAPTIVE:
+			if (openvpn_binary_version >= nmovpn_version_encode (2, 4, 0))
+				_LOGW ("\"comp-lzo\" is deprecated and will be removed in future OpenVPN releases");
+
+			args_add_strv (args, "--comp-lzo",
+			               comp == NMOVPN_COMP_LEGACY_LZO_DISABLED
+			               ? "no"
+			               : "adaptive");
+			break;
+		}
+
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_FLOAT);
 	if (nm_streq0 (tmp, "yes"))
 		args_add_strv (args, "--float");
@@ -1681,6 +1693,8 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 
 	args_add_vpn_data (args, s_vpn, NM_OPENVPN_KEY_CIPHER, "--cipher");
 
+	args_add_vpn_data (args, s_vpn, NM_OPENVPN_KEY_DATA_CIPHERS, "--data-ciphers");
+
 	args_add_vpn_data (args, s_vpn, NM_OPENVPN_KEY_TLS_CIPHER, "--tls-cipher");
 
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_KEYSIZE);
@@ -1724,8 +1738,11 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 	}
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_TLS_VERSION_MIN);
 	if (nmovpn_arg_is_set (tmp)) {
+		const char *or_highest = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_TLS_VERSION_MIN_OR_HIGHEST);
+
 		args_add_strv (args, "--tls-version-min");
-		args_add_strv (args, tmp);
+		args_add_strv0 (args, tmp, nm_streq0(or_highest, "yes") ? "or-highest" : NULL);
+
 	}
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_TLS_VERSION_MAX);
 	if (nmovpn_arg_is_set (tmp)) {
@@ -1741,7 +1758,7 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_TLS_REMOTE);
 	if (nmovpn_arg_is_set (tmp)) {
-		if (openvpn_binary_detect_version_cached (openvpn_binary, &openvpn_binary_version) == OPENVPN_BINARY_VERSION_2_3_OR_OLDER) {
+		if (openvpn_binary_detect_version_cached (openvpn_binary, &openvpn_binary_version) < nmovpn_version_encode (2, 4, 0)) {
 			_LOGW ("the tls-remote option is deprecated and removed from OpenVPN 2.4. Update your connection to use verify-x509-name (for example, \"verify-x509-name=name:%s\")", tmp);
 			args_add_strv (args, "--tls-remote", tmp);
 		} else {
@@ -2312,7 +2329,7 @@ static gboolean
 signal_handler (gpointer user_data)
 {
 	g_main_loop_quit (user_data);
-	return G_SOURCE_REMOVE;
+	return G_SOURCE_CONTINUE;
 }
 
 static void
